﻿// Copyright © 2012 X-treem Software Limited
// Adapted from WPF AutoComplete TextBox Control by Jason Payne
// Licence: http://wpfgooglesearch.codeplex.com/license
// Credit: http://wpfactb.codeplex.com

using Microsoft.Windows.Themes;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Interop;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

namespace Xtreem.GoogleSearch.AutoSuggest
{
    public class AutoSuggestTextBox : TextBox
    {
        #region Constants

        private const int WM_NCLBUTTONDOWN = 0x00A1;
        private const int WM_NCRBUTTONDOWN = 0x00A4;
        private const int POPUP_SHADOW_DEPTH = 5;

        #endregion Constants

        #region Fields

        public static readonly DependencyProperty DataProviderProperty = DependencyProperty.Register(
            "DataProvider",
            typeof(IAutoSuggestDataProvider),
            typeof(AutoSuggestTextBox),
            new UIPropertyMetadata(null));

        private double _itemHeight;
        private double _downWidth;
        private double _downHeight;
        private double _downTop;
        private Point _ptDown;

        private bool _popupOnTop = true;
        private bool _manualResized;
        private string _textBeforeChangedByCode;
        private bool _textChangedByCode;

        private Popup _popup;
        private SystemDropShadowChrome _chrome;
        private ListBox _listBox;
        private ScrollBar _scrollBar;
        private ResizeGrip _resizeGrip;
        private ScrollViewer _scrollViewer;
        private CancellationTokenSource _cancellationTokenSource;

        private bool _disabled;
        private bool _supressAutoAppend;

        #endregion Fields

        #region Constructors

        static AutoSuggestTextBox()
        {
            DefaultStyleKeyProperty.OverrideMetadata(typeof(AutoSuggestTextBox), new FrameworkPropertyMetadata(typeof(AutoSuggestTextBox)));
        }

        public AutoSuggestTextBox()
        {
            if (IsLoaded)
            {
                Initialize();
            }
            else
            {
                Loaded += TextBox_Loaded;
            }
        }

        #endregion Constructors

        #region Properties

        public IAutoSuggestDataProvider DataProvider
        {
            get { return (IAutoSuggestDataProvider)GetValue(DataProviderProperty); }
            set { SetValue(DataProviderProperty, value); }
        }

        public bool Disabled
        {
            get { return _disabled; }
            set
            {
                _disabled = value;
                if (_disabled && _popup != null)
                {
                    _popup.IsOpen = false;
                }
            }
        }

        public bool AutoCompleting
        {
            get { return _popup.IsOpen; }
        }

        public bool Asynchronous { get; set; }

        public bool AutoAppend { get; set; }

        #endregion Properties

        private void Initialize()
        {
            Window.GetWindow(this).LocationChanged += OwnerWindow_LocationChanged;

            _listBox = new ListBox();
            var tempItem = new ListBoxItem { Content = "TEMP_ITEM_FOR_MEASUREMENT" };
            _listBox.Items.Add(tempItem);
            _listBox.Focusable = false;
            _listBox.Style = (Style)Application.Current.Resources["AutoSuggestListBoxStyle"];

            _chrome = new SystemDropShadowChrome();
            _chrome.Margin = new Thickness(0, 0, POPUP_SHADOW_DEPTH, POPUP_SHADOW_DEPTH);
            _chrome.Child = _listBox;

            _popup = new Popup();
            _popup.SnapsToDevicePixels = true;
            _popup.AllowsTransparency = true;
            _popup.MinHeight = SystemParameters.HorizontalScrollBarHeight + POPUP_SHADOW_DEPTH;
            _popup.MinWidth = SystemParameters.VerticalScrollBarWidth + POPUP_SHADOW_DEPTH;
            _popup.VerticalOffset = SystemParameters.PrimaryScreenHeight + 100;
            _popup.Child = _chrome;
            _popup.IsOpen = true;

            _itemHeight = tempItem.ActualHeight;
            _listBox.Items.Clear();

            // Get inner element references
            _scrollViewer = (_listBox.Template.FindName("Border", _listBox) as Border).Child as ScrollViewer;
            _resizeGrip = _scrollViewer.Template.FindName("ResizeGrip", _scrollViewer) as ResizeGrip;
            _scrollBar = _scrollViewer.Template.FindName("PART_VerticalScrollBar", _scrollViewer) as ScrollBar;

            // Update grip visual
            double rectSize = SystemParameters.VerticalScrollBarWidth;
            var triangle = (Path)_resizeGrip.Template.FindName("RG_TRIANGLE", _resizeGrip);
            var pg = (PathGeometry)triangle.Data;
            pg = pg.CloneCurrentValue();
            PathFigure figure = pg.Figures.First();
            Point point = figure.StartPoint;
            point.X = rectSize;
            figure.StartPoint = point;
            var line = (PolyLineSegment)figure.Segments.First();
            point = line.Points[0];
            point.Y = rectSize;
            line.Points[0] = point;
            point = line.Points[1];
            point.X = point.Y = rectSize;
            line.Points[1] = point;
            triangle.Data = pg;

            SetupEventHandlers();
        }

        private void SetupEventHandlers()
        {
            var ownerWindow = Window.GetWindow(this);
            ownerWindow.PreviewMouseDown += OwnerWindow_PreviewMouseDown;
            ownerWindow.Deactivated += OwnerWindow_Deactivated;

            var wih = new WindowInteropHelper(ownerWindow);
            var hwndSource = HwndSource.FromHwnd(wih.Handle);
            var hwndSourceHook = new HwndSourceHook(HookHandler);
            hwndSource.AddHook(hwndSourceHook);
            //hwndSource.RemoveHook();?

            TextChanged += TextBox_TextChanged;
            PreviewKeyDown += TextBox_PreviewKeyDown;

            _listBox.PreviewMouseLeftButtonDown += ListBox_PreviewMouseLeftButtonDown;
            _listBox.MouseLeftButtonUp += ListBox_MouseLeftButtonUp;
            _listBox.PreviewMouseMove += ListBox_PreviewMouseMove;

            _resizeGrip.PreviewMouseLeftButtonDown += ResizeGrip_PreviewMouseLeftButtonDown;
            _resizeGrip.PreviewMouseMove += ResizeGrip_PreviewMouseMove;
            _resizeGrip.PreviewMouseUp += ResizeGrip_PreviewMouseUp;
        }

        #region TextBox Event Handling

        private void TextBox_Loaded(object sender, RoutedEventArgs e)
        {
            Initialize();
        }

        private void TextBox_TextChanged(object sender, TextChangedEventArgs e)
        {
            if (_textChangedByCode || Disabled || DataProvider == null)
            {
                return;
            }

            string text = Text;
            if (string.IsNullOrEmpty(text))
            {
                _popup.IsOpen = false;
                return;
            }

            TaskScheduler uiScheduler = TaskScheduler.FromCurrentSynchronizationContext();

            Action<IEnumerable<TextBlock>> getSuggestionsComplete = delegate(IEnumerable<TextBlock> items)
            {
                PopulatePopupList(items);
            };

            // Prevent race condition should a call be made to this function before another asynchronous call has completed
            if (_cancellationTokenSource != null)
            {
                _cancellationTokenSource.Cancel();
            }

            _cancellationTokenSource = new CancellationTokenSource();
            var cancellationToken = _cancellationTokenSource.Token;
            Task.Factory.StartNew<IEnumerable<TextBlock>>(GetSuggestionsAsync, new Tuple<IAutoSuggestDataProvider, string>(DataProvider, Text), cancellationToken).ContinueWith(x => getSuggestionsComplete(x.Result), new CancellationToken(), TaskContinuationOptions.NotOnCanceled, uiScheduler);
        }

        private IEnumerable<TextBlock> GetSuggestionsAsync(object state)
        {
            var data = (Tuple<IAutoSuggestDataProvider, string>)state;
            return data.Item1.GetSuggestions(data.Item2);
        }

        private void TextBox_PreviewKeyDown(object sender, KeyEventArgs e)
        {
            _supressAutoAppend = e.Key == Key.Delete || e.Key == Key.Back;
            if (!_popup.IsOpen)
            {
                return;
            }

            if (e.Key == Key.Enter)
            {
                _popup.IsOpen = false;
                SelectAll();
            }
            else if (e.Key == Key.Escape)
            {
                _popup.IsOpen = false;
                e.Handled = true;
            }
            if (!_popup.IsOpen)
            {
                return;
            }

            int index = _listBox.SelectedIndex;
            if (e.Key == Key.PageUp)
            {
                if (index == -1)
                {
                    index = _listBox.Items.Count - 1;
                }
                else if (index == 0)
                {
                    index = -1;
                }
                else if (index == _scrollBar.Value)
                {
                    index -= (int)_scrollBar.ViewportSize;
                    if (index < 0)
                    {
                        index = 0;
                    }
                }
                else
                {
                    index = (int)_scrollBar.Value;
                }
            }
            else if (e.Key == Key.PageDown)
            {
                if (index == -1)
                {
                    index = 0;
                }
                else if (index == _listBox.Items.Count - 1)
                {
                    index = -1;
                }
                else if (index == _scrollBar.Value + _scrollBar.ViewportSize - 1)
                {
                    index += (int)_scrollBar.ViewportSize - 1;
                    if (index > _listBox.Items.Count - 1)
                    {
                        index = _listBox.Items.Count - 1;
                    }
                }
                else
                {
                    index = (int)(_scrollBar.Value + _scrollBar.ViewportSize - 1);
                }
            }
            else if (e.Key == Key.Up)
            {
                if (index == -1)
                {
                    index = _listBox.Items.Count - 1;
                }
                else
                {
                    --index;
                }
            }
            else if (e.Key == Key.Down)
            {
                ++index;
            }

            if (index != _listBox.SelectedIndex)
            {
                string text;
                if (index < 0 || index > _listBox.Items.Count - 1)
                {
                    text = _textBeforeChangedByCode;
                    _listBox.SelectedIndex = -1;
                }
                else
                {
                    _listBox.SelectedIndex = index;
                    _listBox.ScrollIntoView(_listBox.SelectedItem);
                    text = _listBox.SelectedItem as string;
                }
                UpdateText(text, false);
                e.Handled = true;
            }
        }

        #endregion TextBox Event Handling

        #region ListBox Event Handling

        private void ListBox_PreviewMouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            var hitTestResult = VisualTreeHelper.HitTest(_listBox, e.GetPosition(_listBox));
            if (hitTestResult != null)
            {
                DependencyObject element = hitTestResult.VisualHit;
                while (element != null)
                {
                    if (element is ListBoxItem)
                    {
                        e.Handled = true;
                        break;
                    }
                    element = VisualTreeHelper.GetParent(element);
                }
            }
        }

        private void ListBox_PreviewMouseMove(object sender, MouseEventArgs e)
        {
            if (Mouse.Captured != null)
            {
                return;
            }

            var hitTestResult = VisualTreeHelper.HitTest(_listBox, e.GetPosition(_listBox));
            if (hitTestResult != null)
            {
                DependencyObject element = hitTestResult.VisualHit;
                while (element != null)
                {
                    var item = element as ListBoxItem;
                    if (item != null)
                    {
                        item.IsSelected = true;
                        //_listBox.ScrollIntoView(item);
                        break;
                    }
                    element = VisualTreeHelper.GetParent(element);
                }
            }
        }

        private void ListBox_MouseLeftButtonUp(object sender, MouseButtonEventArgs e)
        {
            ListBoxItem item = null;
            var d = e.OriginalSource as DependencyObject;
            while (d != null)
            {
                if (d is ListBoxItem)
                {
                    item = d as ListBoxItem;
                    break;
                }
                d = VisualTreeHelper.GetParent(d);
            }
            if (item != null)
            {
                _popup.IsOpen = false;
                UpdateText(item.Content as string, true);
            }
        }

        #endregion ListBox Event Handling

        #region ResizeGrip Event Handling

        private void ResizeGrip_PreviewMouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            _downWidth = _chrome.ActualWidth + POPUP_SHADOW_DEPTH;
            _downHeight = _chrome.ActualHeight + POPUP_SHADOW_DEPTH;
            _downTop = _popup.VerticalOffset;

            var p = e.GetPosition(_resizeGrip);
            p = _resizeGrip.PointToScreen(p);
            _ptDown = p;

            _resizeGrip.CaptureMouse();
        }

        private void ResizeGrip_PreviewMouseMove(object sender, MouseEventArgs e)
        {
            if (e.LeftButton != MouseButtonState.Pressed)
            {
                return;
            }
            var ptMove = e.GetPosition(_resizeGrip);
            ptMove = _resizeGrip.PointToScreen(ptMove);
            var dx = ptMove.X - _ptDown.X;
            var dy = ptMove.Y - _ptDown.Y;
            var newWidth = _downWidth + dx;

            if (newWidth != _popup.Width && newWidth > 0)
            {
                _popup.Width = newWidth;
            }
            if (PopupOnTop)
            {
                var bottom = _downTop + _downHeight;
                var newTop = _downTop + dy;
                if (newTop != _popup.VerticalOffset && newTop < bottom - _popup.MinHeight)
                {
                    _popup.VerticalOffset = newTop;
                    _popup.Height = bottom - newTop;
                }
            }
            else
            {
                var newHeight = _downHeight + dy;
                if (newHeight != _popup.Height && newHeight > 0)
                {
                    _popup.Height = newHeight;
                }
            }
        }

        private void ResizeGrip_PreviewMouseUp(object sender, MouseButtonEventArgs e)
        {
            _resizeGrip.ReleaseMouseCapture();
            if (_popup.Width != _downWidth || _popup.Height != _downHeight)
            {
                _manualResized = true;
            }
        }

        #endregion ResizeGrip Event Handling

        #region Window Event Handling

        private void OwnerWindow_LocationChanged(object sender, EventArgs e)
        {
            _popup.IsOpen = false;
        }

        private void OwnerWindow_Deactivated(object sender, EventArgs e)
        {
            _popup.IsOpen = false;
        }

        private void OwnerWindow_PreviewMouseDown(object sender, MouseButtonEventArgs e)
        {
            if (e.Source != this)
            {
                _popup.IsOpen = false;
            }
        }

        private IntPtr HookHandler(IntPtr hwnd, int msg, IntPtr wParam, IntPtr lParam, ref bool handled)
        {
            handled = false;

            switch (msg)
            {
                case WM_NCLBUTTONDOWN:
                case WM_NCRBUTTONDOWN:
                    _popup.IsOpen = false;
                    break;
            }
            return IntPtr.Zero;
        }

        #endregion Window Event Handling

        #region States and Behaviours

        private void PopulatePopupList(IEnumerable<TextBlock> items)
        {
            string text = Text;

            _listBox.ItemsSource = items;
            if (_listBox.Items.Count == 0)
            {
                _popup.IsOpen = false;
                return;
            }

            string firstSuggestion = ((TextBlock)_listBox.Items[0]).Text;
            if (_listBox.Items.Count == 1 && text.Equals(firstSuggestion, StringComparison.OrdinalIgnoreCase))
            {
                _popup.IsOpen = false;
            }
            else
            {
                _listBox.SelectedIndex = -1;
                _textBeforeChangedByCode = text;
                _scrollViewer.ScrollToHome();
                ShowPopup();

                if (AutoAppend && !_supressAutoAppend && SelectionLength == 0 && SelectionStart == Text.Length)
                {
                    _textChangedByCode = true;
                    try
                    {
                        string appendText = firstSuggestion.Substring(Text.Length);

                        if (!string.IsNullOrEmpty(appendText))
                        {
                            SelectedText = appendText;
                        }
                    }
                    finally
                    {
                        _textChangedByCode = false;
                    }
                }
            }
        }

        private bool PopupOnTop
        {
            get { return _popupOnTop; }
            set
            {
                if (_popupOnTop == value)
                {
                    return;
                }

                _popupOnTop = value;
                if (_popupOnTop)
                {
                    _resizeGrip.VerticalAlignment = VerticalAlignment.Top;
                    _scrollBar.Margin = new Thickness(0, SystemParameters.HorizontalScrollBarHeight, 0, 0);
                    _resizeGrip.LayoutTransform = new ScaleTransform(1, -1);
                    _resizeGrip.Cursor = Cursors.SizeNESW;
                }
                else
                {
                    _resizeGrip.VerticalAlignment = VerticalAlignment.Bottom;
                    _scrollBar.Margin = new Thickness(0, 0, 0, SystemParameters.HorizontalScrollBarHeight);
                    _resizeGrip.LayoutTransform = Transform.Identity;
                    _resizeGrip.Cursor = Cursors.SizeNWSE;
                }
            }
        }

        private void ShowPopup()
        {
            bool popupOnTop = false;

            var p = new Point(0, ActualHeight);
            p = PointToScreen(p);
            double tbBottom = p.Y;

            p = new Point(0, 0);
            p = PointToScreen(p);
            var tbTop = p.Y;

            _popup.HorizontalOffset = p.X;
            double popupTop = tbBottom;

            if (!_manualResized)
            {
                _popup.Width = ActualWidth + POPUP_SHADOW_DEPTH;
            }

            double popupHeight;
            if (_manualResized)
            {
                popupHeight = _popup.Height;
            }
            else
            {
                int visibleCount = Math.Min(16, _listBox.Items.Count + 1);
                popupHeight = visibleCount * _itemHeight + POPUP_SHADOW_DEPTH;
            }
            var screenHeight = SystemParameters.PrimaryScreenHeight;
            if (popupTop + popupHeight > screenHeight)
            {
                if (screenHeight - tbBottom > tbTop)
                {
                    popupHeight = SystemParameters.PrimaryScreenHeight - popupTop;
                }
                else
                {
                    popupOnTop = true;
                    popupTop = tbTop - popupHeight + 4;
                    if (popupTop < 0)
                    {
                        popupTop = 0;
                        popupHeight = tbTop + 4;
                    }
                }
            }

            PopupOnTop = popupOnTop;
            _popup.Height = popupHeight;
            _popup.VerticalOffset = popupTop;

            _popup.IsOpen = true;
        }

        private void UpdateText(string text, bool selectAll)
        {
            _textChangedByCode = true;
            Text = text;

            if (selectAll)
            {
                SelectAll();
            }
            else
            {
                SelectionStart = text.Length;
            }

            _textChangedByCode = false;
        }

        #endregion States and Behaviours
    }
}
